/*
 * Copyright 2014 Garrett D'Amore <garrett@damore.org>
 * Copyright 2010 Nexenta Systems, Inc.  All rights reserved.
 * Copyright (c) 1996 - 2002 FreeBSD Project
 * Copyright (c) 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Paul Borman at Krystal Technologies.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "lint.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <limits.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "collate.h"
#include "lnumeric.h"	/* for struct lc_numeric */
#include "lctype.h"	/* for struct lc_ctype */
#include "setlocale.h"
#include "../i18n/_loc_path.h"
#include "localeimpl.h"
#include "../i18n/_locale.h"

/*
 * Path to locale storage directory.  See ../i18n/_loc_path.h
 */
char	*_PathLocale = _DFLT_LOC_PATH;

static char	*current_locale(locale_t, int);
static void	install_legacy(locale_t, int);

char *
setlocale(int category, const char *locname)
{
	locale_t oldloc;
	locale_t newloc;
	int mask;

	if (category < 0 || category > LC_ALL) {
		errno = EINVAL;
		return (NULL);
	}

	if (locname == NULL)
		return (current_locale(__global_locale, category));

	if ((oldloc = duplocale(__global_locale)) == NULL)
		return (NULL);

	mask = (category == LC_ALL ? LC_ALL_MASK : (1 << category));

	newloc = newlocale(mask, locname, oldloc);
	if (newloc == NULL) {
		freelocale(oldloc);
		return (NULL);
	}

	oldloc = __global_locale;
	__global_locale = newloc;

	install_legacy(newloc, mask);

	freelocale(oldloc);
	return (current_locale(newloc, category));
}

static char *
current_locale(locale_t loc, int cat)
{
	int composite = 0;
	static char buf[LC_ALL * (ENCODING_LEN + 1 + 1)];

	if (cat != LC_ALL) {
		return (loc->locdata[cat]->l_lname);
	}

	/* Look to see if any category is different */
	for (int i = 1; i < LC_ALL; ++i) {
		if (strcmp(loc->locdata[0]->l_lname,
		    loc->locdata[i]->l_lname) != 0) {
			composite = 1;
			break;
		}
	}

	if (composite) {
		/*
		 * Note ordering of these follows the numeric order,
		 * if the order is changed, then setlocale() will need
		 * to be changed as well.
		 */
		(void) snprintf(buf, sizeof (buf),
		    "%s/%s/%s/%s/%s/%s",
		    loc->locdata[LC_CTYPE]->l_lname,
		    loc->locdata[LC_NUMERIC]->l_lname,
		    loc->locdata[LC_TIME]->l_lname,
		    loc->locdata[LC_COLLATE]->l_lname,
		    loc->locdata[LC_MONETARY]->l_lname,
		    loc->locdata[LC_MESSAGES]->l_lname);
		return (buf);
	}

	return (loc->locdata[0]->l_lname);
}

static void
install_legacy(locale_t loc, int mask)
{
	/*
	 * Update the legacy fixed variables that may be baked into
	 * legacy programs.  This is really unfortunate, but we can't
	 * solve for them otherwise.  Note that such legacy programs
	 * are only going to see the global locale settings, and cannot
	 * benefit from uselocale().
	 */
	if (mask & LC_NUMERIC_MASK) {
		struct lc_numeric *lnum;
		lnum = loc->locdata[LC_NUMERIC]->l_data[0];
		_numeric[0] = *lnum->decimal_point;
		_numeric[1] = *lnum->thousands_sep;
	}

	if (mask & LC_CTYPE_MASK) {
		struct lc_ctype *lct;
		lct = loc->locdata[LC_CTYPE]->l_data[0];
		for (int i = 0; i < _CACHED_RUNES; i++) {
			/* ctype can only encode the lower 8 bits. */
			__ctype[i+1] = lct->lc_ctype_mask[i] & 0xff;
			__ctype_mask[i] = lct->lc_ctype_mask[i];
		}

		/* The bottom half is the toupper/lower array */
		for (int i = 0; i < _CACHED_RUNES; i++) {
			int u, l;
			__ctype[258 + i] = i;
			u = lct->lc_trans_upper[i];
			l = lct->lc_trans_lower[i];
			if (u && u != i)
				__ctype[258+i] = u;
			if (l && l != i)
				__ctype[258+i] = l;

			/* Don't forget these annoyances either! */
			__trans_upper[i] = u;
			__trans_lower[i] = l;
		}

		/* Maximum mblen, cswidth, weird legacy */
		__ctype[520] = lct->lc_max_mblen;
	}
}
